package org.netkernel.mod.architecture;

import java.util.*;

import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.impl.HDSFactory;
import org.netkernel.layer0.representation.impl.HDSNodeListImpl;
import org.netkernel.layer0.util.RequestBuilder;
import org.netkernel.layer0.util.RequestBuilder.Arguments;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.ISpace;
import org.netkernel.urii.impl.Version;
import org.w3c.dom.Document;

public class SEConfig implements ISERep
{
	private static class SEConfigElement
	{	public IHDSNode mConfig;
		public ISpace mSpace;
	}
	
	private final List<SEConfigElement> mViews = new ArrayList<SEConfigElement>();
	
	public void addConfig(IHDSNode aConfig, ISpace aSpace)
	{
		SEConfigElement sece=new SEConfigElement();
		sece.mConfig=aConfig;
		sece.mSpace=aSpace;
		mViews.add(sece);
	}
	
	public void close()
	{
	}
	
	public IHDSNodeList getViewsFor(String aEntityType)
	{	List<IHDSNode> nodes = new ArrayList<IHDSNode>();
		for (SEConfigElement sece: mViews)
		{	IHDSNode config=sece.mConfig;
			String entityType=(String)config.getFirstValue("type");
			if (entityType!=null && entityType.equals(aEntityType))
			{	nodes.add(config);
			}
		}
		IHDSNodeList nl=new HDSNodeListImpl(nodes);
		
		Comparator<IHDSNode> c = new Comparator<IHDSNode>()
		{
			public int compare(IHDSNode o1, IHDSNode o2)
			{	String order1=(String)o1.getFirstValue("order");
				String order2=(String)o2.getFirstValue("order");
				if (order1==null) order1="9999";
				if (order2==null) order2="9999";
				return new Integer(order1).compareTo(new Integer(order2));
			}
		};
		nl=nl.sort(c);
		return nl;
	}
	
	private SEConfigElement getConfigFor(String aViewId)
	{
		SEConfigElement found=null;
		for (SEConfigElement sece: mViews)
		{	IHDSNode config=sece.mConfig;
			//System.out.println(config);
			String id=(String)config.getFirstValue("id");
			if (aViewId.equals(id))
			{	found=sece;
				break;
			}
		}
		return found;
	}
	
	public IHDSNode getSource(String aViewId) throws Exception
	{
		SEConfigElement found=getConfigFor(aViewId);
		if (found==null)
		{	throw new Exception("View not found");
		}
		return found.mConfig;
	}
	
	
	public INKFResponseReadOnly generateView(String aViewId, String aEntityId, INKFRequestContext aContext) throws Exception
	{
		SEConfigElement found=getConfigFor(aViewId);
		if (found==null)
		{	throw new Exception("View not found");
		}
		IHDSNode hdsReq=found.mConfig.getFirstNode("request");
		Document req=HDSFactory.toDOM(hdsReq);
		RequestBuilder rb=new RequestBuilder(req, aContext.getKernelContext().getKernel().getLogger());
		RequestBuilder.Arguments args=new RequestBuilder.Arguments();
		if (aEntityId!=null)
		{	args.addArgument("entityId", aEntityId);
		}
		IRequestScopeLevel originalScope=aContext.getKernelContext().getRequestScope();
		IRequestScopeLevel scope=RequestScopeLevelImpl.createNonDurableScopeLevel(found.mSpace, originalScope);
		aContext.getKernelContext().setRequestScope(scope);
		INKFRequest nkfReq=rb.buildRequest(aContext, args, null);
		INKFResponseReadOnly result=aContext.issueRequestForResponse(nkfReq);
		
		aContext.getKernelContext().setRequestScope(originalScope);
		return result;
	}
	
	public Class getClass(String aViewId, INKFRequestContext aContext) throws Exception
	{	SEConfigElement found=getConfigFor(aViewId);
		if (found==null)
		{	throw new Exception("View not found");
		}
		String className=(String)found.mConfig.getFirstValue("class");
		return found.mSpace.getClassLoader().loadClass(className);
		
	}


	
}
