package org.netkernel.mod.architecture.bits;

import java.io.File;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.atomic.AtomicBoolean;

import org.netkernel.ext.system.representation.IRepDeployedModules;
import org.netkernel.layer0.boot.BootUtils;
import org.netkernel.layer0.meta.IEndpointMeta;
import org.netkernel.layer0.meta.IModuleMeta;
import org.netkernel.layer0.meta.IPhysicalEndpointMeta;
import org.netkernel.layer0.meta.IPrototypeMeta;
import org.netkernel.layer0.meta.IRepresentationMeta;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.meta.ISpaceMeta;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.INKFResponseReadOnly;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.urii.IEndpoint;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.IMetaRepresentation;

public class SearchAccessor extends StandardAccessorImpl
{
	private String mIndexDir;
	private HDSBuilder mIndexSpecificationBuilder;
	private int mIndexCount;
	private AtomicBoolean mIndexBuilt=new AtomicBoolean();
	
	public void onSource(INKFRequestContext aContext) throws Exception
	{
		String type=aContext.getThisRequest().getArgumentValue("type");
		if (type.equals("Page"))
		{	onSearchPage(aContext);
		}
		else if (type.equals("Index"))
		{	onSearchIndex(aContext);
		}
		else if (type.equals("Results"))
		{	onSearchResults(aContext);
		}
	}
	
	public void onSearchPage(INKFRequestContext aContext) throws Exception
	{
		lazyInitialiseIndex(aContext,false);
		
		INKFRequest req=aContext.createRequest("active:xrl2");
		req.addArgument("template", "res:/org/netkernel/mod/architecture/bits/searchPage.xml");
		HDSBuilder b=new HDSBuilder();
		INKFResponseReadOnly respIn=aContext.issueRequestForResponse(req);
		INKFResponse resp=aContext.createResponseFrom(respIn);
		resp.setHeader("WrappedControlPanel", Boolean.TRUE);
		
	}
	
	public void onSearchResults(INKFRequestContext aContext) throws Exception
	{
		lazyInitialiseIndex(aContext,true);
		
		String entityId=aContext.getThisRequest().getArgumentValue("entityId");
		IRepDeployedModules modules=aContext.source("active:moduleList", IRepDeployedModules.class);
		
		INKFRequest req=aContext.createRequest("active:textSearch");
		req.addArgument("directory", mIndexDir);
		req.addArgument("limit", "50");
		req.addArgumentByValue("operand", "d1:("+entityId+") d2:("+entityId+") d1:("+entityId+"*)" );
		IHDSNode rs=(IHDSNode)aContext.issueRequest(req);
		
		//boost different entities differently
		IHDSNodeList hits=rs.getNodes("/hits/hit");
		Map<String,Float> scores = new HashMap<String,Float>();
		float maxScore=0.0f;
		for (IHDSNode hit: hits)
		{
			float score=Float.parseFloat((String)hit.getFirstValue("score"));
			String hash=(String)hit.getFirstValue("field[1]/value");
			String type=(String)hit.getFirstValue("field[2]/value");
			float weight=1.0f;
			if (type.equals("M")) weight=4.0f;
			if (type.equals("R")) weight=4.0f;
			if (type.equals("P")) weight=4.0f;
			if (type.equals("S")) weight=1.5f;
			if (type.equals("E")) weight=0.5f;
			score*=weight;
			if (score>maxScore) maxScore=score;
			scores.put(hash,score);
		}
		
		HDSBuilder b=new HDSBuilder();
		String pageId="sr"+org.netkernel.mod.architecture.Utils.getUniqueId();
		b.pushNode("div");
		b.addNode("@id", "explorersearchresults");
		
		StringBuilder script=new StringBuilder();
		script.append("initLinks('#"+pageId+"', '#"+pageId+" .el-physical','gMenuPhysical'); ");
		script.append("initLinks('#"+pageId+"', '#"+pageId+" .el-module','gMenuModule'); ");
		script.append("initLinks('#"+pageId+"', '#"+pageId+" .el-ep','gMenuEp'); ");
		script.append("initLinks('#"+pageId+"', '#"+pageId+" .el-space','gMenuSpace'); ");
		script.append("initLinks('#"+pageId+"', '#"+pageId+" .el-rep','gMenuRep'); ");
		script.append("initLinks('#"+pageId+"', '#"+pageId+" .el-proto','gMenuProto'); ");
		script.append("initDataTable('"+pageId+"', { 'bSort': true, 'aaSorting': [[2,'desc']], 'iDisplayLength': 25 }); ");
		b.pushNode("script",script.toString());
		b.addNode("@type","text/javascript");
		b.popNode();
		
		
		b.pushNode("table");
		b.addNode("@style","width: 100%;");
		b.addNode("@id",pageId);
		b.pushNode("thead");
		b.pushNode("tr");
		b.addNode("th", "Item");
		b.addNode("th", "Module");
		b.addNode("th", "Score");
		b.popNode();
		b.popNode();
		
		
		for (IHDSNode hit: hits)
		{
			b.pushNode("tr");
			String hash=(String)hit.getFirstValue("field/value");
			IMetaRepresentation meta=modules.metadataForHash(hash);
			String icon="";
			String entityClass="";
			if (meta instanceof IRepresentationMeta)
			{	icon="ui-icon-ae ui-icon-ae-rep";
				entityClass="el-rep";
			}
			else if (meta instanceof IPrototypeMeta)
			{	icon="ui-icon-ae ui-icon-ae-proto";
				entityClass="el-proto";
			}
			else if (meta instanceof IEndpointMeta)
			{	icon="ui-icon-ae ui-icon-ae-ep";
				entityClass="el-ep";
			}
			else if (meta instanceof ISpaceMeta)
			{	icon="ui-icon-ae ui-icon-ae-space";
				entityClass="el-space";
			}
			else if (meta instanceof IModuleMeta)
			{	icon="ui-icon-ae ui-icon-ae-module";
				entityClass="el-module";
			}
			else if (meta instanceof IPhysicalEndpointMeta)
			{	icon="ui-icon-ae ui-icon-ae-physical";
				entityClass="el-physical";
			}
			b.pushNode("td");
			b.addNode("@entity", hash);
			b.addNode("@title", meta.getDescription());
			b.addNode("@class","el "+entityClass);
			b.pushNode("span");
			b.addNode("@class", icon);
			b.popNode();
			b.addNode("span", meta.getName());
			b.popNode();

			int modi=hash.indexOf('/');
			String modHash;
			if (modi>0)
			{	modHash=hash.substring(0,modi);
			}
			else
			{	modHash=hash;
			}
			IMetaRepresentation modMeta=modules.metadataForHash(modHash);
			b.pushNode("td");
			b.addNode("@entity", modHash);
			b.addNode("@class","el el-module");
			b.pushNode("span");
			b.addNode("@class", "ui-icon-ae ui-icon-ae-module");
			b.popNode();
			b.addNode("span", modMeta.getName());
			b.popNode();

			float score=scores.get(hash);
			score/=maxScore;
			
			b.pushNode("td");
			b.pushNode("div");
			
			b.pushNode("span",score);
			b.addNode("@style", "display: none;");
			b.popNode();
			
			b.addNode("style", "margin-top: 5px; margin-right: 5px; float: right;");
			b.pushNode("span");
			b.addNode("@style", "background: #219b3f;");
			b.pushNode("img");
			b.addNode("@src","/nkse/style/img/blank.png");
			b.addNode("@width",score*50.0f);
			b.addNode("@height","16");
			b.popNode();
			b.popNode();
			b.pushNode("span");
			b.addNode("@style", "background: white;");
			b.pushNode("img");
			b.addNode("@src","/nkse/style/img/blank.png");
			b.addNode("@width",(1-score)*50.0f);
			b.addNode("@height","16");
			b.popNode();
			b.popNode();

			b.popNode();
			b.popNode();
			
		}
		aContext.createResponseFrom(b.getRoot());
	}
	
	private void lazyInitialiseIndex(INKFRequestContext aContext, boolean aWait) throws Exception
	{	if (mIndexDir==null)
		{	
			INKFRequest req=aContext.createRequest("res:/aebits/view/searchIndex");
			if (aWait)
			{	aContext.issueRequest(req);
			}
			else
			{	aContext.issueAsyncRequest(req);
			}
		}
		else if (!mIndexBuilt.get() && aWait)
		{	while (!mIndexBuilt.get())
			{	Thread.sleep(50);
			}
		}
	}
	
	public void onSearchIndex(INKFRequestContext aContext) throws Exception
	{
		index(aContext);
	}
	
	private void index(INKFRequestContext aContext) throws Exception
	{
		String indexDirName="ExplorerSearchIndex"+System.currentTimeMillis();
		File temp=File.createTempFile(indexDirName, null);
		File tempDir=temp.getParentFile();
		temp.delete();
		File searchDirFile=new File(tempDir, indexDirName+File.pathSeparator);
		searchDirFile.mkdir();
		String searchDir=searchDirFile.toURI().toString();
		mIndexDir=searchDir;
		resetIndex();
		
		IRepDeployedModules modules=aContext.source("active:moduleList", IRepDeployedModules.class);
		IHDSNode list=modules.getHierarchy();
		for (IHDSNode module: list.getNodes("/modules/module"))
		{	String mHash=(String)module.getFirstValue("hash");
			IModuleMeta moduleMeta=(IModuleMeta)modules.metadataForHash(mHash);
			indexItem(moduleMeta,mHash,aContext);
			for (IHDSNode space: module.getNodes("spaces//space"))
			{	String spaceHash=(String)space.getFirstValue("hash");
				ISpaceMeta spaceMeta=(ISpaceMeta)modules.metadataForHash(spaceHash);
				indexItem(spaceMeta,spaceHash, aContext);
				ISpaceElements se=spaceMeta.getElements();
				IEndpoint lastEp=null;
				for (int i=0; i<se.size(); i++)
				{	IEndpoint ep=se.getPhysicalEndpoint(i);
					if (ep!=lastEp)
					{	lastEp=ep;
						String epHash=modules.hashForComponent(ep);
						IMetaRepresentation epMeta=modules.metadataForHash(epHash);
						indexItem(epMeta,epHash,aContext);
					}
					
					if (se.getDelegatedSpace(i)==null)
					{	IIdentifier eid=se.getIdentifier(i);
						if (eid!=null)
						{	String eHash=spaceHash+"/"+eid.toString();
							IMetaRepresentation eMeta=modules.metadataForHash(eHash);
							indexItem(eMeta,eHash,aContext);
						}
					}
				}
			}
		}
		writeIndex(aContext);
		mIndexBuilt.set(true);
	}
	
	private void resetIndex()
	{	mIndexCount=0;
		mIndexSpecificationBuilder=new HDSBuilder();
		mIndexSpecificationBuilder.pushNode("batch");
	}
	
	private String indexItem(IMetaRepresentation aMeta,String aHash, INKFRequestContext aContext) throws Exception
	{
		StringBuilder sb1=new StringBuilder(128);
		StringBuilder sb2=new StringBuilder(128);

		String rawName=aMeta.getName();
		sb1.append(rawName);
		sb1.append(' ');
		String camelName=unCamelize(rawName);
		//if (!rawName.equals(camelName))
		{	sb2.append(camelName);
			sb2.append(' ');
		}
		if (aMeta.getDescription()!=null)
		{	sb1.append(aMeta.getDescription());
			sb1.append(' ');
		}
		String type="";
		if (aMeta instanceof IModuleMeta)
		{	sb1.append(((IModuleMeta)aMeta).getSource().toString());
			type="M";
		}
		else if (aMeta instanceof IPhysicalEndpointMeta)
		{	type="E";
		}
		else if (aMeta instanceof ISpaceMeta)
		{	type="S";
		}
		else if (aMeta instanceof IRepresentationMeta)
		{	
			String className=((IRepresentationMeta)aMeta).getImplementation().getName();
			className=className.replaceAll("\\.", " ");
			sb1.append(className);
			type="R";
		}
		else if (aMeta instanceof IPrototypeMeta)
		{	String className=((IPrototypeMeta)aMeta).getImplementation().getName();
			className=className.replaceAll("\\.", " ");
			sb1.append(className);
			sb1.append(' ');
			sb1.append(aHash);
			type="P";
		}
		else if (aMeta instanceof IEndpointMeta)
		{
			String rawGrammar=((IEndpointMeta)aMeta).getIdentifierGrammar().toString();
			sb1.append(rawGrammar);
			sb1.append(' ');
			String grammar=unCamelize(rawGrammar);
			//if (!grammar.equals(rawGrammar))
			{	sb2.append(grammar);
			}
			sb1.append(' ');
			sb1.append(aHash);
			type="L";
		}
		mIndexSpecificationBuilder.pushNode("item");
		addIndexField("hash",aHash,"NO","YES");
		addIndexField("type",type,"NO","YES");
		addIndexField("d1",sb1.toString(),"ANALYZED","NO");
		addIndexField("d2",sb2.toString(),"ANALYZED","NO");
		mIndexSpecificationBuilder.popNode();
		mIndexCount++;
		
		if (mIndexCount>399)
		{
			writeIndex(aContext);
			resetIndex();
		}
		
		
		return sb1.toString();
	}
	
	private void addIndexField(String aName, String aValue, String aAnalyze, String aStore)
	{
		mIndexSpecificationBuilder.pushNode("field");
		mIndexSpecificationBuilder.addNode("@index", aAnalyze);
		mIndexSpecificationBuilder.addNode("@store", aStore);
		mIndexSpecificationBuilder.addNode("name", aName);
		mIndexSpecificationBuilder.addNode("value", aValue);
		mIndexSpecificationBuilder.popNode();
	}
	
	private void writeIndex(INKFRequestContext aContext) throws Exception
	{	INKFRequest req=aContext.createRequest("active:textIndexBatch");
		req.addArgument("directory", mIndexDir);
		req.addArgumentByValue("operand", mIndexSpecificationBuilder.getRoot());
		aContext.issueRequest(req);
	}
	
	private static String unCamelize(String aInput)
	{
		//transition from lower->upper or char->number or number->char causes split
		StringBuilder result=new StringBuilder();
		boolean lastUpper=false;
		boolean lastDigit=false;
		for (int i=0; i<aInput.length(); i++)
		{
			char c=aInput.charAt(i);
			boolean upper=Character.isUpperCase(c);
			boolean digit=Character.isDigit(c);
			if (i>0)
			{
				if ( (upper && !lastUpper) || (digit!=lastDigit) )
				{	result.append(" ");
				}
				if (lastUpper && upper && i<aInput.length()-1 && Character.isLowerCase(aInput.charAt(i+1)))
				{	result.append(" ");
				}
			}
			result.append(c);
			lastUpper=upper;
			lastDigit=digit;
		}
		return result.toString();
	}	
	
}
