package org.netkernel.mod.architecture.bits.staticstructure;

import java.awt.geom.Rectangle2D;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.text.NumberFormat;
import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;

import org.netkernel.ext.system.representation.IRepDeployedModules;
import org.netkernel.ext.system.representation.impl.SpaceGraph;
import org.netkernel.ext.system.representation.impl.SpaceGraph.Edge;
import org.netkernel.layer0.boot.IModule;
import org.netkernel.layer0.meta.IEndpointStateMeta;
import org.netkernel.layer0.meta.ISpaceMeta;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.representation.impl.HDSPredicateFactory;
import org.netkernel.layer0.util.MultiMap;
import org.netkernel.layer0.util.Utils;
import org.netkernel.mod.architecture.ISERep;
import org.netkernel.module.standard.endpoint.IStandardEndpoint;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.module.standard.endpoint.StandardTransportImpl;
import org.netkernel.urii.IEndpoint;
import org.netkernel.urii.IMetaRepresentation;
import org.netkernel.urii.ISpace;

public class StaticStructureDiagramAccessor extends StandardAccessorImpl
{
	public final static float ENDPOINT_HEIGHT=9;
	public final static float XSCALE=170.0f;
	public final static float XOFFSET=100.0f;
	public final static float YSCALE=170.0f;
	public final static float MODULE_TITLE_HEIGHT=40.0f;

	private static AtomicInteger sUniqueId=new AtomicInteger();
	private final NumberFormat mNF;
	
	public StaticStructureDiagramAccessor()
	{	DecimalFormat df=new DecimalFormat("0.00");
		DecimalFormatSymbols dfs=df.getDecimalFormatSymbols();
		dfs.setDecimalSeparator('.');
		df.setDecimalFormatSymbols(dfs);
		mNF=df;
	}
	
	public void onSource(INKFRequestContext aContext) throws Exception
	{
		IRepDeployedModules modules = aContext.source("active:moduleList", IRepDeployedModules.class);
		SpaceGraph sg = aContext.source("active:moduleListSpaceGraph", SpaceGraph.class);

		IHDSNode params = aContext.source("arg:params", IHDSNode.class);
		List<String> roots = new ArrayList<String>();
		for (IHDSNode param : params.getChildren())
		{
			if (param.getName().equals("root"))
			{
				roots.add((String) param.getValue());
			}
		}
		
		int iter;
		try
		{	iter=Integer.parseInt(aContext.source("httpRequest:/param/iter",String.class));
		}	catch (Exception e) {iter=800; }
		float scale;
		try
		{	scale=Float.parseFloat(aContext.source("httpRequest:/param/scale",String.class));
		}	catch (Exception e) {scale=1.0f; }
		boolean wrap=aContext.source("httpRequest:/param/wrap",String.class)!=null;

		String decoratorId = (String) params.getFirstValue("decorator");
		ISERep config = aContext.source("active:seConfig",ISERep.class);
		IStaticStructureDiagramDecorator decorator;
		if (decoratorId!=null)
		{	decorator=(IStaticStructureDiagramDecorator)config.getClass(decoratorId,aContext).newInstance();
		}
		else
		{	decorator=new DefaultDecorator();
		}
		IHDSNode settings=aContext.source("fpds:/etc/system/explorerSettings.xml",IHDSNode.class);
		decorator.initialise(params, modules, sg, settings);
		
		Map<String,GraphNode> nodeMap=new HashMap<String, GraphNode>();
		List<GraphNode> nodes2=new ArrayList<GraphNode>();
		Set<GraphEdge> edges=new HashSet<GraphEdge>();
		Random r=new Random(0);
		
		float totalh=0;
		for (int i=0; i<roots.size(); i++)
		{	String root=roots.get(i);
			GraphNode n=recurseSpace2(root,modules,sg,0,0,nodeMap,nodes2,edges,null,null,r,false,0,decorator,aContext);
			int h=n.getYSpan();
			float y=totalh*YSCALE;
			recursiveOffsetY(n, null, y, nodeMap, sg, modules, new HashSet<GraphNode>());
			totalh+=h;
		}
		
		anneal2(nodes2,edges,iter,decorator);
		
		IHDSNode s=dump(nodes2,edges,modules,roots,scale,wrap, decorator,aContext, sg);
		INKFResponse resp=aContext.createResponseFrom(s);
		resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
		if (wrap)
		{	resp.setMimeType("text/html");
		}
		else
		{	resp.setMimeType("image/svg+xml");
		}
		
		if (params.getFirstValue("download")!=null)
		{	resp.setHeader("httpResponse:/header/Content-Disposition", "attachment; filename=ssd.svg");
		}
		
		
	}
	
	private GraphNode recurseSpace2(String aHash, IRepDeployedModules aModules, SpaceGraph aGraph, int aDepth, int aLayoutDepth, Map<String,GraphNode> aNodesByHash, List<GraphNode> aNodes, Set<GraphEdge> aEdges, GraphNode aParent, String aParentEndpointHash, Random aR, boolean aParentNew, float aXTweak, IStaticStructureDiagramDecorator aDecorator, INKFRequestContext aContext)
	{
		ISpace space=(ISpace)aModules.componentForHash(aHash);
		String name=space.toString();
		{
			boolean system=aDecorator.isLibrarySpace(aHash);
			GraphNode n;
			
			if (aDecorator.mergeLibraries())
			{	n=aNodesByHash.get(aHash);
			}
			else
			{
				if (system)
				{	if (aParentNew)
					{	n=null;
					}
					else
					{	n=null;
						for (GraphNode n2:aParent.mEdgesOut)
						{	if (n2.mHash.equals(aHash))
							{	n=n2;
								break;
							}
						}
					}
				}
				else
				{	n=aNodesByHash.get(aHash);
				}
			}
			
			boolean exists=(n!=null);
			if (exists)
			{	
				if (aLayoutDepth>n.mDepth)
				{	n.mDepth=aLayoutDepth;
					n.mX=XOFFSET+aLayoutDepth*XSCALE;
				}
			}
			else
			{	float x;
				if (aParent==null)
				{	x=XOFFSET;
				}
				else
				{	x=XOFFSET+aLayoutDepth*XSCALE; 
					x+= (aR.nextFloat()-0.5f)*5;
				}

				n=new GraphNode(aHash,x,system,aModules,aGraph);
				n.mDepth=aLayoutDepth;
				float y=aParent!=null?aParent.mY:0.0f;
				n.updateYPos(y, aParentEndpointHash);
				aNodesByHash.put(aHash, n);
				aNodes.add(n);
			}

			GraphEdge edgeIn=null;
			if (aParent!=null && (!exists || aParentNew))
			{	aParent.addNodeOut(n);
				n.addNodeIn(aParent);
				
				
				Set<String> injectedSpaces = getInjectedSpaces(aParentEndpointHash, aParent.mHash, aModules, aGraph, aContext);
				
				GraphEdge edge=new GraphEdge(aParent,n,aParentEndpointHash,injectedSpaces);
				edgeIn=edge;
				aEdges.add(edge);
				aParent.mNEdgesOut.add(edge);
				if (aParentNew && exists)
				{	n.updateYPos(aParent.mY, aParentEndpointHash);
				}
			}
			
			Edge[] edges=aGraph.getEdgesOut(aHash);
			int edgeCount=edges.length;
			int[] edgesHeight = new int[edgeCount];
			boolean[] edgeEnabled = new boolean[edgeCount];
			
			int enabledEdgeCount=0;
			for (int i=0; i<edgeCount; i++)
			{	boolean enabled=aDecorator.showEdge(n.mHash,edges[i].getSpaceHash(),aDepth+1);
				if (enabled)
				{	edgeEnabled[i]=true;
					enabledEdgeCount++;
				}
			}
			
			int totalHeight=0;
			int xTweak=0;
			if (enabledEdgeCount>2)
			{	xTweak=1;	
			}
			
			for (int i=0; i<edgeCount; i++)
			{	Edge edge=edges[i];
				String spaceHash=edge.getSpaceHash();
				String epHash=edge.getPhysicalEndpointHash();
				
				Set<String> injected=getInjectedSpaces(epHash, aHash, aModules, aGraph, aContext);
				int xTweak2=0;
				if (injected.size()>0)
				{	xTweak2=1;
				}
				
				if (edgeEnabled[i])
				{	recurseSpace2(spaceHash,aModules,aGraph, aDepth+1, n.mDepth+1+xTweak+xTweak2, aNodesByHash,aNodes,aEdges,n,epHash,aR,!exists,0,aDecorator,aContext);
				}
			}
			
			int nedgeCount=n.mNEdgesOut.size();
			int[] edgesHeight2 = new int[nedgeCount];
			int th2=0;
			for (int i=0; i<nedgeCount; i++)
			{
				GraphEdge nedge=n.mNEdgesOut.get(i);
				int h=nedge.mNode2.getYSpan();
				edgesHeight2[i]=h;
				th2+=h;
			}
			
			if (!exists)
			{
				float th=0;
				for (int i=0; i<nedgeCount; i++)
				{
					int h=edgesHeight2[i];
					GraphEdge ne=n.mNEdgesOut.get(i);
					
					float s=(float)th2;
					float ix=th+((float)h)*0.5f;
					float j=(ix-s*0.5f);
					float separation=(ne.mNode2.mEdgesOut.size()==0)?YSCALE/2.0f:YSCALE;
					float y=j*separation;
					//System.out.println("OFFSET: "+ne.mNode2.mHash+" "+y);
					recursiveOffsetY(ne.mNode2, ne.mEndpointHash, y, aNodesByHash, aGraph, aModules,new HashSet<GraphNode>());
					th+=h;
				}
			}
			
			
			if (totalHeight==0)
			{	totalHeight=1;
			}
			//System.out.println(totalHeight+" "+aModules.componentForHash(n.mHash));
			
			
			//NEED a better test to see if this node is shared- this will only detect second use but not first use of shared
			//two pass, 
			if (exists)
			{	totalHeight=-totalHeight;
			}
			return n;
			
		}
	}
	
	private static Set<String> getInjectedSpaces(String aEndpointHash, String aSpaceHash, IRepDeployedModules aModules, SpaceGraph aGraph, INKFRequestContext aContext)
	{
		Set<String> injectedSpaces = new HashSet<String>();
		IEndpoint physicalEp=(IEndpoint)aModules.componentForHash(aEndpointHash);
		if (physicalEp instanceof IStandardEndpoint)
		{	try
			{	ISpace[] spaces=((IStandardEndpoint)physicalEp).getSpaces(aContext);
				for (ISpace epSpace: spaces)
				{	
					String epSpaceHash=aModules.hashForComponent(epSpace);
					injectedSpaces.add(epSpaceHash);
				}
			}
			catch (Exception e)
			{
			}
		}
		for (Edge e: aGraph.getEdgesOut(aSpaceHash))
		{	String spHash=e.getSpaceHash();
			injectedSpaces.remove(spHash);					
		}
		return injectedSpaces;
	}
	
	private static void recursiveOffsetY(GraphNode aNode, String aParentHash, float aY, Map<String,GraphNode> aNodesByHash, SpaceGraph aGraph, IRepDeployedModules aModules, Set<GraphNode> aNodesOffset)
	{
		if (aY==0.0f) return;
		aNodesOffset.add(aNode);
		float ydiff=aNode.mY;
		int s0=aNode.mYComponentsByParent.size();
		Float yo=aNode.mYComponentsByParent.get(aParentHash);
		float y;
		if (yo==null)
		{	y=aY;
		}
		else
		{	y=yo+aY;
		}
		aNode.updateYPos(y, aParentHash);
		ydiff=(aNode.mY-ydiff);
		
		for (GraphEdge edge: aNode.mNEdgesOut)
		{	recursiveOffsetY(edge.mNode2, edge.mEndpointHash, ydiff, aNodesByHash, aGraph, aModules,aNodesOffset);
		}
	}
	
	private static void anneal2(List<GraphNode> aNodes, Set<GraphEdge> aEdges, int aIter, IStaticStructureDiagramDecorator aDecorator)
	{
		final float SPACING_SAME=Math.min(XSCALE, YSCALE)*0.5f;
		final float SPACING_DIFF=(float)Math.sqrt(XSCALE*XSCALE+YSCALE*YSCALE)*(aDecorator.showModules()?0.7f:0.5f);
		
		for (int i=0; i<aIter; i++)
		{
			for (GraphNode node1:aNodes)
			{
				float fX=0.0f,fY=0.0f;
				int j=node1.mHash.indexOf('/');
				String node1ModuleHash=node1.mHash.substring(0,j);
				
				//push apart if too close
				for (GraphNode node2: aNodes)
				{	if (node2!=node1)
					{
						j=node2.mHash.indexOf('/');
						String node2ModuleHash=node2.mHash.substring(0,j);
						float sep=(node1ModuleHash.equals(node2ModuleHash))?SPACING_SAME:SPACING_DIFF;
					
						float dx=node2.mX-node1.mX;
						float dy=node2.mY-node1.mY;
						float d2=dx*dx+dy*dy;
						float d=(float)Math.sqrt(d2);
						dx/=d;
						dy/=d;
						float f=d>=sep?0:-20.0f;
						//if (f<-1) f=-1;
						
						fX+=f*dx;
						fY+=f*dy;					
					}
				}
				
				//push apart from injected spaces too (hack)
				for (GraphEdge edge: aEdges)
				{	
					if (edge.mInjectedSpaces.size()>0)
					{
						
						float x3=edge.mNode1.mX + 50;
						int ie=edge.mNode1.indexOfEndpoint(edge.mEndpointHash);
						float y3=edge.mNode1.mY +edge.mNode1.getYPosOfEndpoint(ie);
						
						float x0=edge.mNode2.mX - 63;
						float y0=edge.mNode2.mY;
						
						float R3=4;
						float T3=40;
						
						//String spaceHash=edge.mInjectedSpaces.iterator().next();
						float ix=x3+(x0-x3)/2;
						float iy=y3+(y0-y3)/2+R3*2+T3;
						
						float dx=ix-node1.mX;
						float dy=iy-node1.mY;
						float d2=dx*dx+dy*dy;
						float d=(float)Math.sqrt(d2);
						dx/=d;
						dy/=d;
						float f=d>=SPACING_SAME?0:-20.0f;
						//if (f<-1) f=-1;
						
						fX+=f*dx;
						fY+=f*dy;					
						
					}
				}
				
				

				//preserve left right orientation
				List<GraphNode> nl=new ArrayList<GraphNode>(node1.mEdgesIn.size()+node1.mEdgesOut.size());
				nl.addAll(node1.mEdgesIn);
				nl.addAll(node1.mEdgesOut);
				
				for (int k=0; k<nl.size(); k++)
				{
					GraphNode node2=nl.get(k);
					float dx=node2.mX-node1.mX;
					boolean dir=k<node1.mEdgesIn.size();
					float df=dir?1:-1;
					boolean wrongWay=dir!=((dx-XSCALE*0.9f)<0); //XSCALE
					if (wrongWay)
					{	fX+=df*20f;
						//System.out.println("wrong way");
					}
					
				}
				
				node1.mXV=fX;
				node1.mYV=fY;
			}
			
			if (aDecorator.showModules())
			{
				float maxd=(float)Math.sqrt(XSCALE*XSCALE + YSCALE*YSCALE)*1.0f;
				MultiMap poolSet=pool(aNodes,maxd);
				
				//we have more than one pool from same module than might be pulled together
				for (Iterator it=poolSet.keyIterator(); it.hasNext();)
				{	String moduleHash=(String)it.next();
					List<Set<GraphNode>> pools=poolSet.get(moduleHash);
					if (pools.size()>0)
					{
						
						for (Set<GraphNode> pool1: pools)
						{
							for (Set<GraphNode> pool2: pools)
							{
								if (pool1!=pool2)
								{
									for (GraphNode node1: pool1)
									{
										//int lc=node1.mEdgesIn.size()+node1.mEdgesOut.size();
										float lc=pool1.size();
										for (GraphNode node2: pool2)
										{
											if (node1.mEdgesIn.contains(node2) || node1.mEdgesOut.contains(node2))
											{
												float dx=node2.mX-node1.mX;
												float dy=node2.mY-node1.mY;
												float d2=dx*dx+dy*dy;
												float d=(float)Math.sqrt(d2);
												dx/=d;
												dy/=d;
												float f=(d/40.0f)/(lc*lc);
												//System.out.println("REPEL "+d+" "+lc+" "+f);
												node1.mXV+=f*dx;
												node1.mYV+=f*dy;		
												
											}
										}
									}
								}
							}
						}
						
					}
				}
				
				//look for intersecting pools from different modules and push apart
				List<String> moduleHashes = new ArrayList<String>();
				for (Iterator it=poolSet.keyIterator(); it.hasNext();)
				{	moduleHashes.add((String)it.next());
				}
				for (int k=1; k<moduleHashes.size(); k++)
				{	String module1Hash=moduleHashes.get(k);
					List<Set<GraphNode>> module1Pools=poolSet.get(module1Hash);
					for (Set<GraphNode> module1Pool : module1Pools)
					{
						Rectangle2D.Float module1Rec=getBoundingRectangle(module1Pool);
						module1Rec.y-=MODULE_TITLE_HEIGHT;
						module1Rec.height+=MODULE_TITLE_HEIGHT;

						for (int j=0; j<k; j++)
						{	String module2Hash=moduleHashes.get(j);
							if (!module1Hash.equals(module2Hash))
							{
								List<Set<GraphNode>> module2Pools=poolSet.get(module2Hash);
								for (Set<GraphNode> module2Pool : module2Pools)
								{
									Rectangle2D.Float module2Rec=getBoundingRectangle(module2Pool);
									module2Rec.y-=MODULE_TITLE_HEIGHT;
									module2Rec.height+=MODULE_TITLE_HEIGHT;
									
									if (module2Rec.intersects(module1Rec))
									{
										//System.out.println(module1Hash+" intersects "+module2Hash);
										
										//determine a direction N S E W to move
										Rectangle2D.Float is=new Rectangle2D.Float();
										Rectangle2D.Float.intersect(module2Rec, module2Rec, is);
										boolean moveX=is.width<is.height;
										float fx=0f;
										float fy=0f;
										float F=10.0f;
										if (moveX)
										{	float dx=(module2Rec.x+module2Rec.width)-(module1Rec.x+module2Rec.width);
											fx=(dx>0)?-F:F;
										}
										else
										{	float dy=(module2Rec.y+module2Rec.height)-(module1Rec.y+module2Rec.height);
											fy=(dy>0)?-F:F;
										}
										
										for (GraphNode node: module1Pool)
										{	node.mXV+=fx;
											node.mYV+=fy;
										}
										for (GraphNode node: module2Pool)
										{	node.mXV-=fx;
											node.mYV-=fy;
										}
									}
											
								}
							}
						}
					}
				}
			}
			
			boolean energy=false;
			for (GraphNode node1: aNodes)
			{	//update positions based on velocity
				node1.mX+=1.0f*node1.mXV;
				node1.mY+=1.0f*node1.mYV;
				energy|=(node1.mXV!=0.0f || node1.mYV!=0.0f);
			}
			if (!energy)
			{	//System.out.println("break after "+i);
				break;
			}
		}
	}
	
	private static MultiMap pool(List<GraphNode> aNodes, float aMaxDistance)
	{
		float maxd2=aMaxDistance*aMaxDistance;
		MultiMap poolSet=new MultiMap(100, 2);
		
		for (GraphNode node1: aNodes)
		{
			int io=node1.mHash.indexOf('/');
			String node1ModuleHash=node1.mHash.substring(0,io);
			List<Set> poolsForModule=poolSet.get(node1ModuleHash);
			List<Set<GraphNode>> foundPools=new ArrayList();
			for (Set<GraphNode> pool: poolsForModule)
			{	for (GraphNode node2: pool)
				{
					float dx=node2.mX-node1.mX;
					float dy=node2.mY-node1.mY;
					float d2=dx*dx+dy*dy;
					if (d2<maxd2)
					{	foundPools.add(pool);
						break;
					}
				}
			}
			if (foundPools.size()==0)
			{	Set<GraphNode> foundPool = new HashSet<GraphNode>();
				poolSet.put(node1ModuleHash, foundPool);
				foundPool.add(node1);
			}
			else if (foundPools.size()==1)
			{	Set<GraphNode> foundPool = foundPools.get(0);
				foundPool.add(node1);
			}
			else
			{	//need to merge pools
				Set<GraphNode> mergedPool = new HashSet<GraphNode>();
				mergedPool.add(node1);
				for (Set<GraphNode> pool: foundPools)
				{	mergedPool.addAll(pool);
					poolSet.remove(node1ModuleHash,pool);
				}
				poolSet.put(node1ModuleHash, mergedPool);
			}
				
		}
		return poolSet;
	}
	
	private static Rectangle2D.Float getBoundingRectangle(Set<GraphNode> aPool)
	{
		float xmax=Float.NEGATIVE_INFINITY;
		float xmin=Float.POSITIVE_INFINITY;
		float ymax=Float.NEGATIVE_INFINITY;
		float ymin=Float.POSITIVE_INFINITY;
		for (GraphNode node: aPool)
		{	if (xmax<node.mX) xmax=node.mX;
			if (xmin>node.mX) xmin=node.mX;
			float h2=node.getHeight()/2;
			if (ymax<node.mY+h2) ymax=node.mY+h2;
			if (ymin>node.mY-h2) ymin=node.mY-h2;
		}
		xmax+=XSCALE/2;
		xmin-=XSCALE/2;
		ymax+=YSCALE/4;
		ymin-=YSCALE/4;
		return new Rectangle2D.Float(xmin,ymin,xmax-xmin,ymax-ymin);
	}
	
	
	private IHDSNode dump(List<GraphNode> aNodes, Set<GraphEdge> aEdges, IRepDeployedModules aModules, List<String> aRoots, float aScale, boolean aWrap, IStaticStructureDiagramDecorator aDecorator, INKFRequestContext aContext, SpaceGraph sg) throws NKFException
	{
		//float R=0.0f;
		//float T1=8.0f;
		//float T2=6.0f;
		
		//determine bounds
		float maxX=Float.MIN_VALUE;
		float maxY=Float.MIN_VALUE;
		float minX=Float.MAX_VALUE;
		float minY=Float.MAX_VALUE;
		for (GraphNode node: aNodes)
		{	
			float x=node.mX;
			float y=node.mY;
			float h=node.getHeight()/2;
			if (maxX<x) maxX=x;
			if (maxY<y+h) maxY=y+h;
			if (minX>x) minX=x;
			if (minY>y-h) minY=y-h;
		}
		final float PADDING_X=aDecorator.showModules()?XSCALE*0.6f:XSCALE*0.5f;
		final float PADDING_Y=aDecorator.showModules()?YSCALE*0.6f:YSCALE*0.5f;
		maxX+=PADDING_X;
		maxY+=PADDING_Y;
		minX-=PADDING_X;
		minY-=PADDING_Y;
		float w=(maxX-minX)*aScale;
		float h=(maxY-minY)*aScale;
		
		StringBuilder sb=new StringBuilder();		
		HDSBuilder b=new HDSBuilder();
		
		if (aWrap)
		{
			b.pushNode("html");
			b.pushNode("head");
			b.pushNode("script","_");
			b.addNode("@data-path", "/nkse/style/svgweb/svg.js");
			b.addNode("@src", "/nkse/style/svgweb/svg.js");
			b.popNode();
			/*
			b.pushNode("link");
			b.addNode("@type", "text/css");
			b.addNode("@href", "/se3/pub/staticstructurediagram.css");
			b.popNode();
			*/
			b.popNode();
			b.pushNode("div");
			b.addNode("@style", "width: "+(int)w+"px; height: "+(int)h+"px; background: #EEE;");
			b.pushNode("script");
			b.addNode("@type", "image/svg+xml");
		}
		
		b.pushNode("svg");
		b.addNode("@xmlns","http://www.w3.org/2000/svg");
		b.addNode("@version", "1.1");
		b.addNode("@width",(int)w);
		b.addNode("@height",(int)h);
		
		b.pushNode("style", aContext.source("res:/org/netkernel/mod/architecture/bits/staticstructure/staticstructurediagram.css",String.class));
		b.addNode("@type","text/css");
		b.popNode();
				
		String id="fga"+sUniqueId.getAndIncrement();
		
		//space drop shadow
		b.pushNode("filter");
		b.addNode("@id", id+"-i1");
		b.addNode("@x","-5%");
		b.addNode("@y","-5%");
		//b.addNode("@width","150%");
		//b.addNode("@height","150%");
		b.pushNode("feOffset");
		b.addNode("@result","offOut");
		b.addNode("@in","SourceAlpha");
		b.addNode("@dx",2);
		b.addNode("@dy",2);
		b.popNode();
		b.pushNode("feGaussianBlur");
		b.addNode("@result","blurOut");
		b.addNode("@in","offOut");
		b.addNode("@stdDeviation",3);
		b.popNode();
		b.pushNode("feBlend");
		b.addNode("@in","SourceGraphic");
		b.addNode("@in2","blurOut");
		b.addNode("@mode","normal");
		b.popNode();
		b.popNode();
		
		//space title outer glow
		b.pushNode("filter");
		b.addNode("@id", id+"-i2");
		//b.addNode("@width","150%");
		//b.addNode("@height","150%");
		b.pushNode("feMorphology");
		b.addNode("@result","dilateOut");
		b.addNode("@in","SourceAlpha");
		b.addNode("@operator","dilate");
		b.addNode("@radius",1);
		b.popNode();
		b.pushNode("feGaussianBlur");
		b.addNode("@result","blurOut");
		b.addNode("@in","dilateOut");
		b.addNode("@stdDeviation",1);
		b.popNode();
		b.pushNode("feColorMatrix");
		b.addNode("@in","blurOut");
		b.addNode("@type","matrix");
		b.addNode("@values","0 0 0 0 1   0 0 0 0 1   0 0 0 0 1   0 0 0 0.75 0");
		b.popNode();
		b.popNode();
		
		//module drop shadow
		b.pushNode("filter");
		b.addNode("@id", id+"-i4");
		b.addNode("@x","-5%");
		b.addNode("@y","-5%");
		//b.addNode("@width","150%");
		//b.addNode("@height","150%");
		b.pushNode("feOffset");
		b.addNode("@result","offOut");
		b.addNode("@in","SourceAlpha");
		b.addNode("@dx",2);
		b.addNode("@dy",2);
		b.popNode();
		b.pushNode("feGaussianBlur");
		b.addNode("@result","result1");
		b.addNode("@in","offOut");
		b.addNode("@stdDeviation",3);
		b.popNode();
		b.pushNode("feColorMatrix");
		b.addNode("@result","result2");
		b.addNode("@in","SourceAlpha");
		b.addNode("@type","matrix");
		b.addNode("@values","0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 -1.5 1");
		b.popNode();
		b.pushNode("feComposite");
		//b.addNode("@result","result3");
		b.addNode("@in","result1");
		b.addNode("@in2","result2");
		b.addNode("@operator","arithmetic");
		b.addNode("@k1","1.0");
		b.addNode("@k2","0.0");
		b.addNode("@k3","0.0");
		b.addNode("@k4","0.0");
		b.popNode();
		/*
		b.pushNode("feBlend");
		b.addNode("@in","SourceGraphic");
		b.addNode("@in2","result3");
		b.addNode("@mode","normal");
		b.popNode();
		*/
		b.popNode();
		
		b.pushNode("g");
		b.addNode("@transform", "scale("+aScale+" "+aScale+")  translate("+(-minX)+", "+(-minY)+")");

		float maxd=(float)Math.sqrt(XSCALE*XSCALE + YSCALE*YSCALE)*1.0f;
		
		if (aDecorator.showModules())
		{	MultiMap poolSet=pool(aNodes, maxd);
			writeModules(b,poolSet,aModules,id,aDecorator);
		}

		for (GraphNode node: aNodes)
		{	writeSpace(b,node,aModules,aEdges,aRoots,id,false,aDecorator);			
		}
		
		for (GraphEdge edge: aEdges)
		{	writeEdge(b, edge, aModules, aDecorator,aContext);
		}
		
		for (GraphEdge edge: aEdges)
		{	
			if (edge.mInjectedSpaces.size()>0)
			{
				float T1=8.0f;
				float T2=6.0f;
				float R=4;
				float W=100;
				
				float x3=edge.mNode1.mX + 50;
				int i=edge.mNode1.indexOfEndpoint(edge.mEndpointHash);
				float y3=edge.mNode1.mY +edge.mNode1.getYPosOfEndpoint(i);
				
				float x0=edge.mNode2.mX - 63;
				float y0=edge.mNode2.mY;
				
				float x1=x0-T1;
				float y1=y0+T2;
				
				float x2=x0-T1;
				float y2=y0-T2;
				
				float R3=4;
				float T3=40;
				
				boolean shade=aDecorator.shade();
				b.pushNode("g");
				b.addNode("@class",(shade?"cl":"cl-noshade"));				
				
				b.pushNode("ellipse");
				b.addNode("@cx",x3+(x0-x3)/2);
				b.addNode("@cy",y3+(y0-y3)/2);
				b.addNode("@rx",R3);
				b.addNode("@ry",R3*2);
				b.popNode();

				b.pushNode("line");
				b.addNode("@x1",x3+(x0-x3)/2);
				b.addNode("@y1",y3+(y0-y3)/2+R3*2);
				b.addNode("@x2",x3+(x0-x3)/2);
				b.addNode("@y2",y3+(y0-y3)/2+R3*2+T3);
				b.popNode();
				
				b.popNode();
				
				String spaceHash=edge.mInjectedSpaces.iterator().next();
				GraphNode n=new GraphNode(spaceHash, x3+(x0-x3)/2, false, aModules, sg);
				n.mY=y3+(y0-y3)/2+R3*2+T3;
				writeSpace( b, n, aModules, aEdges, aRoots, id, false, aDecorator);				
				writeSpace( b, n, aModules, aEdges, aRoots, id, true, aDecorator);				
				
				
			}
		}

		for (GraphNode node: aNodes)
		{	writeSpace(b,node,aModules,aEdges,aRoots,id,true,aDecorator);			
		}

		return b.getRoot();
	}
	
	
	private void writeEdge(HDSBuilder b, GraphEdge edge, IRepDeployedModules aModules, IStaticStructureDiagramDecorator aDecorator, INKFRequestContext aContext)
	{
		boolean shade=aDecorator.shade();
		float T1=8.0f;
		float T2=6.0f;
		float R=4;
		float W=100;
		
		float x3=edge.mNode1.mX + 50;
		int i=edge.mNode1.indexOfEndpoint(edge.mEndpointHash);
		float y3=edge.mNode1.mY +edge.mNode1.getYPosOfEndpoint(i);
		
		float x0=edge.mNode2.mX - 63;
		float y0=edge.mNode2.mY;
		
		float x1=x0-T1;
		float y1=y0+T2;
		
		float x2=x0-T1;
		float y2=y0-T2;
		
		float cpo=((x0-x3)<100)?Math.min(100.0f,Math.abs(y0-y3)):(x0-x3)/2.0f; //control point offset
		
		String epHash=edge.mEndpointHash;
		b.pushNode("g");
			b.addNode("@class",(shade?"cl":"cl-noshade")+" cmPhysical");
			b.addNode("@entity",epHash);
		/*
		b.pushNode("path");
			b.addNode("@d", "M "+x3+" "+y3+" c "+cpo/2.0f+" "+0+" "+(x0-x3-cpo)+" "+(y0-y3)+" "+(x0-x3-T1/2)+" "+(y0-y3));
			b.addNode("@class", "clunder");
		b.popNode();
		*/
		
		b.pushNode("rect");
		b.addNode("@x",x3-100);
		b.addNode("@y",y3-4);
		b.addNode("@width",W-R);
		b.addNode("@height",ENDPOINT_HEIGHT-1);
		b.addNode("@class","cl-startep1");
		b.popNode();
		
		b.pushNode("path");
		b.addNode("@class", "cl-startep2");		
		b.addNode("@d", "M "+(x3-R)+","+(y3-R)+" A "+R+","+R+" 180 0,1 "+(x3-R)+","+(y3+R)+" Z");
		b.popNode();

		b.pushNode("path");
			b.addNode("@d", "M "+(x3)+" "+y3+" c "+cpo+" "+0+" "+(x0-x3-cpo)+" "+(y0-y3)+" "+(x0-x3-T1/2)+" "+(y0-y3));
		b.popNode();
		b.pushNode("polygon");
			b.addNode("@points",x0+","+y0+","+x1+","+y1+","+x2+","+y2);
		b.popNode();
			
		aDecorator.decorateEdge(b,edge);
		b.popNode();
		
	}
	
	private void writeSpace(HDSBuilder b, GraphNode node, IRepDeployedModules aModules, Set<GraphEdge> aEdges, List<String> aRoots, String aId, boolean aFull, IStaticStructureDiagramDecorator aDecorator)
	{
		float w=100.0f;
		float R=4;
		float h=node.getHeight();
		String spaceHash=node.mHash;
		ISpaceMeta spaceMeta=(ISpaceMeta)aModules.metadataForHash(spaceHash);
		
		
		IHDSNodeList spaces=aModules.getHierarchy().getNodes("/modules/module//space");
		IHDSNode sn=spaces.filter(HDSPredicateFactory.namedChildStringEquals("hash", spaceHash)).getFirstNode();
		boolean problems=sn!=null && sn.getFirstNode(IEndpointStateMeta.KEY_PROBLEM)!=null;
		//System.out.println(spaceHash+" "+problems);
		
		boolean shade=aDecorator.shade();
		
		String colour;
		if (!shade)
		{	colour="nspace";
		}
		else if (problems && false)
		{	colour="pspace";
		}
		else if (aRoots.contains(spaceHash))
		{	colour="rspace";
		}
		else if (node.mExternal)
		{	colour="lspace";
		}
		else
		{	colour="space";
		}
		
		b.pushNode("g");
		b.addNode("@transform", "translate("+mNF.format(node.mX-w/2)+" "+mNF.format(node.mY-h/2)+")");
		b.pushNode("g");
		if (!aFull && shade) b.addNode("@filter", "url(#"+aId+"-i1)");
		
		if (!aFull || shade)
		{
			b.pushNode("path");
			b.addNode("@d", getSpacePath(w, h));
			b.addNode("@class",colour+" cmSpace"); b.addNode("@entity",spaceHash); 
			b.addNode("@style","fill-opacity: "+((aFull)?0.5f:1.0f));
			b.popNode();
		}
		b.popNode();
		
		if (aFull)
		{
			String title1=spaceMeta.getName();
			int ix=spaceHash.indexOf('/');
			String moduleHash=spaceHash.substring(0,ix);
			IModule mod=(IModule)aModules.componentForHash(moduleHash);
			ISpace ourSpace=(ISpace)aModules.componentForHash(spaceHash);
			ISpace[] pubSpaces=mod.getPublicSpaces();
			boolean isPublic=false;
			for (ISpace space: pubSpaces)
			{	if (space==ourSpace)
				{	isPublic=true;
					break;
				}
			}
			if (isPublic)
			{	title1="+"+title1;
				if (title1.endsWith("(public)"))
				{	title1=title1.substring(0,title1.length()-8);
				}
			}
			
			// title
			List<String> titleRows=Utils.splitIntoRows(title1, 24, " /-");
			if (shade)
			{
				b.pushNode("g");
				b.addNode("@filter", "url(#"+aId+"-i2");
				for (int i=0; i<titleRows.size(); i++)
				{
					b.pushNode("text",titleRows.get(i));
					//b.addNode("@class","cmSpace"); b.addNode("@entity",spaceHash);
					b.addNode("@x",w/2);
					b.addNode("@y",8+(i-titleRows.size()+1)*10);
					b.addNode("@font-size",10);
					b.addNode("@font-weight","bold");
					b.addNode("@text-anchor","middle");
					b.popNode();
				}			
				b.popNode();
			}
			for (int i=0; i<titleRows.size(); i++)
			{
				b.pushNode("text",titleRows.get(i));
				b.addNode("@class","cmSpace"); b.addNode("@entity",spaceHash);
				b.addNode("@x",w/2);
				b.addNode("@y",8+(i-titleRows.size()+1)*10);
				b.addNode("@font-size",10);
				b.addNode("@font-weight","bold");
				b.addNode("@text-anchor","middle");
				b.popNode();
			}
			/*
			String spaceTitle=node.mHash; //"Hash="+node.mHash+" Y="+node.mY+" YSpan="+node.getYSpan()+" -> "+node.mYComponentsByParent;
			b.pushNode("g");
			b.addNode("@filter", "url(#"+aId+"-i2");
			b.pushNode("text",title1);
			b.addNode("@x",w/2);
			b.addNode("@y",8);
			b.addNode("@font-size",10);
			b.addNode("@font-weight","bold");
			b.addNode("@text-anchor","middle");
			b.popNode();
			b.popNode();
			b.pushNode("text",title1);
			b.addNode("@class","cmSpace"); b.addNode("@entity",spaceHash);
			b.addNode("@x",w/2);
			b.addNode("@y",8);
			b.addNode("@font-size",10);
			b.addNode("@font-weight","bold");
			b.addNode("@text-anchor","middle");
			b.addNode("@title",spaceTitle);
			b.popNode();
			*/
			
			
			
			for (int i=0; i<node.mPhysicalEndpoints.size(); i++)
			{
				String epHash=node.mPhysicalEndpoints.get(i);
				//boolean linked=node.mLinkedEndpoints.get(i);
				boolean linked=false;
				for (GraphEdge nedge: node.mNEdgesOut)
				{	if (nedge.mEndpointHash.equals(epHash))
					{	linked=true;
						break;
					}
				}
				IMetaRepresentation meta=aModules.metadataForHash(epHash);
				Object isTransportObj=meta.getAdditionalFields().getValue(StandardTransportImpl.STATE_TRANSPORT_INDICATOR);
				boolean isTransport=isTransportObj instanceof Boolean && ((Boolean)isTransportObj).booleanValue();

				String cls=linked?"eplinked":"ep";
				if (isTransport) 
				{	drawTransport(b, node,i,aDecorator);
					cls="eplinked";
				} 
				if (!shade)
				{	cls="ep-noshade";
				}

				float y=node.getYPosOfEndpoint(i)+h/2;
				b.pushNode("rect");
				b.addNode("@x",0);
				b.addNode("@y",y-4);
				b.addNode("@width",w-R);
				b.addNode("@height",ENDPOINT_HEIGHT-1);
				b.addNode("@class",cls);
				b.popNode();
				if (linked)
				{
					b.pushNode("path");
					b.addNode("@class", cls);		
					b.addNode("@d", "M "+(w-R)+","+(y-R)+" A "+R+","+R+" 180 0,1 "+(w-R)+","+(y+R)+" Z");
					b.popNode();
				}
				
				IEndpoint ep=(IEndpoint)aModules.componentForHash(epHash);
				String title=Utils.splitIntoRows(ep.toString(), 24, " /-").get(0);		//Utils.shortClassName(ep.getClass());//+" "+epHash;
				String title2=ep.toString();
				b.pushNode("text",title);
				b.addNode("@x",w-R-1);
				b.addNode("@y",y+2);
				b.addNode("@font-size",8);
				b.addNode("@text-anchor","end");
				b.addNode("@title",title2);
				b.popNode();
			
			}
		}
		b.popNode();
	}
	
	private String getSpacePath(float aWidth, float aHeight)
	{
		double[] sc = new double[] { 366.76,99.9, 51.4,-100.3, -65.545,-96.68, -106.42,-75.9, -41.2,-33.5, -96.88,-30.96, -117.58,2.54, -32.7,-30.4, -133.0,-34.7, -113.1,55.7, -41.2,22.2, -40.3,107.79, 10.3,133.6, -31.16,89.575, 56.33,101.7, 100.13,66.9, 12.675,38.845, 112.875,59.1, 134.075,-4.6, 51.4,33.4, 155.83,-17.5, 91.53,-66.9, 46.09,-31.695, 62.67,-89, 1.07,-111.3, 0,0 };
		for (int i=0; i<sc.length; i+=2)
		{	
			if (i<2)
			{	sc[i]= (sc[i]-40.5f);
				sc[i+1]= (sc[i+1]-27.0f);
			}
			sc[i]= sc[i]*aWidth/323.0f;
			sc[i+1]= sc[i+1]*aHeight/248.0f;
		}
		StringBuilder sb=new StringBuilder();
		sb.append("m "+sc[0]+","+sc[1]+" c ");
		for (int i=2; i<sc.length-2; i+=2)
		{	sb.append(mNF.format(sc[i]));
			sb.append(',');
			sb.append(mNF.format(sc[i+1]));
			sb.append(' ');
		}
		int l=sc.length;
		sb.append("l "+sc[l-2]+","+sc[l-1]+" z");		
		return sb.toString();
	}
	
	private void drawTransport(HDSBuilder b, GraphNode aNode, int aEndpointIndex,IStaticStructureDiagramDecorator aDecorator)
	{
		float x1= 0;
		float y1= aNode.getHeight()/2 + aNode.getYPosOfEndpoint(aEndpointIndex);
		
		float ydiff=(aEndpointIndex>=aNode.mPhysicalEndpoints.size()/2)?50:-50;
		
		float x2= - 25;
		float y2=y1+ydiff;
		
		float T1=8.0f;
		float T2=6.0f;
		
		boolean shade=aDecorator.shade();
		b.pushNode("g");
			b.addNode("@class",(shade?"cl":"cl-noshade")+" cmPhysical");
		b.pushNode("path");
			b.addNode("@d", "M "+x2+" "+y2+" c "+0+" "+(-ydiff)+" "+(x1-x2-25)+" "+(y1-y2)+" "+(x1-x2-T1/2)+" "+(y1-y2));
			b.addNode("@class", "clunder");
		b.popNode();
		b.pushNode("circle");
			b.addNode("@cx",mNF.format(x2));
			b.addNode("@cy",mNF.format(y2));
			b.addNode("@r",6);
		b.popNode();
		b.pushNode("path");
			b.addNode("@d", "M "+x2+" "+y2+" c "+0+" "+(-ydiff)+" "+(x1-x2-25)+" "+(y1-y2)+" "+(x1-x2-T1/2)+" "+(y1-y2));
		b.popNode();
		b.pushNode("polygon");
			b.addNode("@points",x1+","+y1+","+(x1-T1)+","+(y1+T2)+","+(x1-T1)+","+(y1-T2));
		b.popNode();
		b.popNode();		
	}	
	
	private static void writeModules(HDSBuilder b, MultiMap aPoolSet, IRepDeployedModules aModules, String aId, IStaticStructureDiagramDecorator aDecorator)
	{
		for (Iterator it=aPoolSet.keyIterator(); it.hasNext();)
		{	String moduleHash=(String)it.next();
			List<Set<GraphNode>> pools=aPoolSet.get(moduleHash);
			for (Set<GraphNode> pool: pools)
			{
				if (pool.iterator().next().mExternal) continue; // don't give external modules a marking
				
				Rectangle2D.Float rect=getBoundingRectangle(pool);
				IMetaRepresentation meta=aModules.metadataForHash(moduleHash);
				
				String title=meta.getName(); //+" ("+moduleHash+")";
				int textSize=10;
				int width=textSize;
				for (int i=0; i<title.length(); i++)
				{	char c=title.charAt(i);
					width+=Character.isLowerCase(c)?textSize:textSize*3/2;
				}
				float height=22;
				
				b.pushNode("g");
				boolean shade=aDecorator.shade();
				if (shade)
				{	b.addNode("@filter", "url(#"+aId+"-i4");
				
					b.pushNode("rect");
					b.addNode("@x",rect.x);
					b.addNode("@y",rect.y);
					b.addNode("@width",rect.width);
					b.addNode("@height",rect.height);
					b.addNode("@class",shade?"module":"module-noshade");
					b.addNode("@title",title);
					b.popNode();
					
					b.pushNode("rect");
					b.addNode("@x",rect.x);
					b.addNode("@y",rect.y-height);
					b.addNode("@width",width);
					b.addNode("@height",height);
					b.addNode("@class",shade?"module":"module-noshade");
					b.addNode("@title",title);
					b.popNode();
					
					b.popNode();
				}
				
				b.pushNode("rect");
				b.addNode("@x",rect.x);
				b.addNode("@y",rect.y);
				b.addNode("@width",rect.width);
				b.addNode("@height",rect.height);
				b.addNode("@class",shade?"module":"module-noshade");
				b.addNode("@title",title);
				b.popNode();
				
				b.pushNode("rect");
				b.addNode("@x",rect.x);
				b.addNode("@y",rect.y-height);
				b.addNode("@width",width);
				b.addNode("@height",height);
				b.addNode("@class",shade?"module":"module-noshade");
				b.addNode("@title",title);
				b.popNode();
				
				
				
				
				b.pushNode("text",title);
				b.addNode("@class","cmModule"); b.addNode("@entity",moduleHash);
				b.addNode("@x",rect.x+textSize/2);
				b.addNode("@y",rect.y-height/4);
				b.addNode("@font-size",15);
				b.addNode("@font-weight","bold");
				b.addNode("@text-anchor","left");
				//b.addNode("@title",title);
				b.popNode();
								
			}
		}
		
	}
}
